const WebSocket = require("ws");
const TCP = require("net");
const UDP = require("dgram");
class InstantiatedSDK {
    constructor(socket_type = "ws") {
        this.s_type=socket_type;
        switch(this.s_type) {
            case "ws":
                this.socket = new WebSocket("wss://instantiated.xyz:12701");
                break;
            case "tcp":
                this.socket = new TCP.Socket();
                this.socket.connect(12702, "instantiated.xyz", function() {});
                break;
            case "udp":
                this.socket = new UDP.createSocket("udp4");
                break;
        }
        this.Session = "";
        this.Socket = -1;
        this.Role = "";
        this.Clients = [];
        this.IsConnected = false;
        this.IsHost = false;
    }

    process(data) {
        switch(this.s_type) {
            case "ws":
                this.socket.send(data);
                break;
            case "tcp":
                this.socket.write(data+"\r\n");
                break;
            case "udp":
                this.socket.send(data, 12700, "instantiated.xyz");
                break;
        }
    }

    onOpen(func) {
        // Sets an onopen event to the this.socket.
        if(this.s_type == "ws") {
            this.socket.on("open", func);
        }
    }

    onMessage(func) {
        // Sets an onmessage event to the this.socket.
        if(this.s_type == "ws" || this.s_type == "udp") {
            this.socket.on("message", func);
        } else {
            this.socket.on("data", func);
        }
    }

    onClose(func) {
        // Sets an onclose event to the this.socket.
        this.socket.on("close", func);
    }

    onError(func) {
        // Sets an onerror event to the this.socket.
        this.socket.on("error", func);
    }

    SetRole(RoleID) {
        this.Role = RoleID;
    }

    GetRole() {
        return this.Role;
    }

    SetSessionID(InstanceID) {
        // Sets the session id to the api.
        this.Session = InstanceID;
    }

    SetSocketID(SocketID) {
        // Sets the socket id to the api.
        this.Socket = SocketID;
    }

    GetSocketID() {
        // Returns the socket id.
        return this.Socket;
    }

    GetSessionID() {
        // Returns the session id.
        return this.Session;
    }

    CreateSession(AppTitle, MaxConnections, Whitelist, Password, RootPassword, DefaultValues, ForceHostTransfer, KillHostServer, AnnounceOpenSession = false) {
        // Creates a new session for the client. The client is now considered the host.
        this.process(JSON.stringify({
            msg: "create_server_session",
            Title: AppTitle,
            max_clients: MaxConnections,
            whitelist: Whitelist,
            passkey: Password,
            root_passkey: RootPassword,
            default_values: DefaultValues,
            force_host_transfer: ForceHostTransfer,
            host_killserver: KillHostServer,
            announceOpenSession: AnnounceOpenSession
        }));
    }

    CreateServer(Session_ID, AppTitle, MaxConnections, Whitelist, Password, DefaultValues, AnnounceOpenSession = false) {
        // Creates a server only based session for the client. The client is now considered the host. Useful if your game is hosting under a console and doesn't require graphics.
        this.process(JSON.stringify({
            msg: "server_only",
            SessionID: Session_ID,
            Title: AppTitle,
            max_clients: MaxConnections,
            whitelist: Whitelist,
            passkey: Password,
            default_values: DefaultValues,
            announceOpenSession: AnnounceOpenSession
        }));
    }

    Disconnect() {
        // Disconnects from the session.
        this.process(JSON.stringify({
            msg: "disconnect",
            instance_guid: this.Session
        }));
    }

    Send(Json) {
        // Sends JSON data or a string to the host.
        this.process(JSON.stringify({
            msg: "sent",
            instance_guid: this.Session,
            DataStream: Json,
            role_id: this.Role
        }));
    }

    Receive(Json) {
        // Sends JSON data or a string to all the clients connected to this session.
        this.process(JSON.stringify({
            msg: "received",
            instance_guid: this.Session,
            DataStream: Json
        }));
    }

    PrivateSend (Json, Socket_ID) {
        // Privately sends JSON data or a string to a single client connected to this session.
        this.process(JSON.stringify({
            msg: "private_msg",
            instance_guid: this.Session,
            DataStream: Json,
            socket: Socket_ID
        }));
    }

    SocketSend (Json, Socket_ID) {
        // Sends JSON Data or a string to a socket that is connected to the Instantiated server. They do not need to be connected to the session to send data to you.
        this.process(JSON.stringify({
            msg: "socket_to",
            DataStream: Json,
            SocketID: Socket_ID
        }));
    }

    Ping () {
        // Pings the network letting them know that you are still alive.
        this.process(`{"msg":"ping"}`);
    }

    TransferOwnership (socket_id) {
        // Transfers ownership to someone else if they're online.
        this.process(JSON.stringify({
            msg: "transfer_ownership",
            instance_guid: this.Session,
            socketid: socket_id,
            role_id: this.Role
        }));
    }

    CreateInvite(max_use) {
        // Creates an invite code for you to send to your friends.
        this.process(JSON.stringify({
            msg: "CreateInvite",
            instance_guid: this.Session,
            limitUse: max_use,
            role_id: this.Role
        }));
    }

    DeleteInvite(code) {
        // Deletes an invite code and invalidates it.
        this.process(JSON.stringify({
            msg: "DeleteInvite",
            instance_guid: this.Session,
            invite_code: code,
            role_id: this.Role
        }));
    }

    AccountExport() {
        // Allows you to backup accounts from the session.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "Export"
        }));
    }

    AccountImport(account_data) {
        // Allows you to restore accounts from json data.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "Import",
            account_json: account_data
        }));
    }

    AccountLogin(username, password) {
        // Logs you in to your registered account in the session.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "Login",
            userid: username,
            passkey: password
        }));
    }

    AccountRegister(username, password, alias) {
        // Registers an account in a connected session.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "Register",
            userid: username,
            passkey: password,
            nickname: alias
        }));
    }

    AccountLogout() {
        // Logs you out of the account.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "Logout"
        }));
    }

    AccountSaveData(Item, Data) {
        // Stores private data that only the account has access to.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "SaveData",
            item: Item,
            data: Data
        }));
    }

    AccountLoadData(Item) {
        // Grabs the account's item data.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "LoadData",
            item: Item
        }));
    }

    AccountDeleteData(Item) {
        // Deletes the account's item.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "DeleteData",
            item: Item
        }));
    }

    AccountReadMail(mail_id, username = null) {
        // Reads messages sent by other accounts.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "ReadMailbox",
            position: mail_id,
            userid: username
        }));
    }

    AccountSendMail(To, Subject, TextMessage) {
        // Sends a message to other accounts.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "SendMailbox",
            to: To,
            subject: Subject,
            textmessage: TextMessage
        }));
    }

    AccountDeleteMessage(Position, Username) {
        // Deletes a message from your account.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "DeleteMessageMail",
            position: Position,
            userid: Username
        }));
    }

    AccountEdit(Alias, Bio, Status, ProfileIMG_Url, Username = null) {
        // Edits the account.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "EditAccount",
            nickname: Alias,
            bio: Bio,
            status: Status,
            profile_img: ProfileIMG_Url,
            userid: Username
        }));
    }

    AccountProfileSearch(Username) {
        // Search for an account using the username result.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "ProfileSearch",
            userid: Username
        }));
    }

    AccountInfo(Username = null) {
        // Checks the account information.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "GrabAccountInfo",
            userid: Username
        }));
    }

    AccountChangePassword(NewPassword, Username = null) {
        // Change the account's password.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "EditAccountPassword",
            passkey: NewPassword,
            userid: Username
        }));
    }

    AccountDelete(Username = null) {
        // Completely deletes the account from the instance.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "DeleteInstanceAccount",
            userid: Username
        }));
    }

    AccountAssignRole(Username, RoleName) {
        // Assigns the account to a role so when logged in they are assigned automatically to that role.
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "AssignAccountRole",
            userid: Username,
            role_name: RoleName
        }));
    }

    AccountUnassignRole(Username, RoleName) {
        // Unassign the account 
        this.process(JSON.stringify({
            msg: "AccountSystem",
            instance_guid: this.Session,
            cmd: "RemoveAssignedAccountRole",
            userid: Username,
            role_name: RoleName
        }));
    }

    ChangeInstancePassword(NewPassword) {
        // Changes the instance's password.
        this.process(JSON.stringify({
            msg: "ChangeInstancePassword",
            instance_guid: this.Session,
            passkey: NewPassword,
            role_id: this.Role
        }));
    }

    AnnounceOnDiscord(Message, Type = "maintenance") {
        // Announces your events to discord.
        this.process(JSON.stringify({
            msg: "AnnounceToDiscord",
            instance_guid: this.Session,
            announcement: Message,
            type: Type,
            role_id: this.Role
        }))
    }

    CloudLogin(user, pass) {
        this.process(JSON.stringify({
            msg: "Cloud",
            action: "Login",
            username: user,
            password: pass
        }));
    }

    CloudRegister(user, pass) {
        // Registers your ip address to the cloud.
        this.process(JSON.stringify({
            msg: "Cloud",
            action: "Register",
            username: user,
            password: pass
        }));
    }

    CloudAdd(Key, Value) {
        // Adds the key data to the cloud.
        this.process(JSON.stringify({
            msg: "Cloud",
            action: "Add",
            Item: Key,
            Data: Value
        }));
    }

    CloudUpdate(Key, Value) {
        // Updates the key data to a value.
        this.process(JSON.stringify({
            msg: "Cloud",
            action: "Update",
            Item: Key,
            Data: Value
        }));
    }

    CloudRemove(Key) {
        // Completely removes the key from the cloud storage.
        this.process(JSON.stringify({
            msg: "Cloud",
            action: "Remove",
            Item: Key
        }));
    }

    CloudAppend(Key, Value) {
        // Add/Appends the data from the cloud storage.
        this.process(JSON.stringify({
            msg: "Cloud",
            action: "Append",
            Item: Key,
            Data: Value
        }));
    }

    CloudGetValue(Key) {
        // Retrieves the cloud key data.
        this.process(JSON.stringify({
            msg: "Cloud",
            action: "Get",
            Item: Key
        }));
    }

    CloudReset() {
        // Resets the storage to zero. Cannot be undone.
        this.process(JSON.stringify({
            msg: "Cloud",
            action: "Reset"
        }));
    }

    CloudInfo() {
        // Displays information about your cloud storage.
        this.process(JSON.stringify({
            msg: "Cloud",
            action: "Info"
        }));
    }

    CloudAccountDelete() {
        this.process(JSON.stringify({
            msg: "Cloud",
            action: "DeleteUserAccount"
        }));
    }

    MailRegister(Username, Password) {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "Register",
            username: Username,
            password: Password
        }));
    }
    
    MailLogin(Username, Password) {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "Login",
            username: Username,
            password: Password
        }));
    }

    MailListFolders() {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "ListFolders"
        }));
    }

    MailListMailInFolder(FolderName) {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "ListMailInFolder",
            folder: FolderName
        }));
    }

    ReadMail(FolderName, MailNumber) {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "ReadMail",
            folder: FolderName,
            mail_id: MailNumber
        }));
    }

    ComposeMail(To, Subj, Msg, invite = null) {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "Compose",
            to: To,
            Subject: Subj,
            Message: Msg,
            InviteCode: invite
        }));
    }

    MarkMailAsRead(FolderName, MailNumber) {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "MarkAsRead",
            folder: FolderName,
            mail_id: MailNumber
        }));
    }

    MarkMailAsUnread(FolderName, MailNumber) {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "MarkAsUnread",
            folder: FolderName,
            mail_id: MailNumber
        }));
    }

    DeleteMail(FolderName, MailNumber) {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "DeleteMail",
            folder: FolderName,
            mail_id: MailNumber
        }));
    }

    MoveMailToFolder(OldFolder, MailNumber, NewFolder) {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "MoveMail",
            oldfolder: OldFolder,
            oldmail: MailNumber,
            newfolder: NewFolder
        }));
    }

    CreateMailFolder(NewFolder) {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "NewFolder",
            folder_name: NewFolder
        }));
    }

    AddSpammer(Username) {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "AddSpammer",
            spammer: Username
        }));
    }

    RemoveSpammer(Username) {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "RemoveSpammer",
            spammer: Username
        }));
    }

    BlockMailUser(Username) {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "BlockUser",
            username: Username
        }));
    }

    UnblockMailUser(Username) {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "UnblockUser",
            username: Username
        }));
    }

    DeleteMailFolder(FolderName) {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "DeleteFolder",
            folder_name: FolderName
        }));
    }

    DeleteMailAccount() {
        this.process(JSON.stringify({
            msg: "Mail",
            action: "DeleteAccount"
        }));
    }

    ExposeClientsIP() {
        // Exposes the clients ip address (instance host and roles that have this permission can use it.)
        this.process(JSON.stringify({
            msg: "ExposeConnectedClients",
            instance_guid: this.Session,
            role_id: this.Role
        }));
    }

    Check_Newsletters() {
        // Displays the latest news from instantiated.
        this.process(JSON.stringify({
            msg: "Newsletters"
        }));
    }

    SetInstanceMOTD(Message) {
        // Sets the message of the day text to something like "A WSTunnel Server."
        this.process(JSON.stringify({
            msg: "MOTD",
            mode: "set",
            instance_guid: this.Session,
            motd: Message,
            role_id: this.Role
        }));
    }

    GetInstanceMOTD() {
        // Retrieves the MOTD message.
        this.process(JSON.stringify({
            msg: "MOTD",
            mode: "get",
            instance_guid: this.Session
        }));
    }

    AddInstanceRule(Rule_String) {
        // Adds your rule to the instance. (Host & Roles with the permissions can add rules.)
        this.process(JSON.stringify({
            msg: "InstanceRules",
            mode: "Add",
            instance_guid: this.Session,
            rule: Rule_String,
            role_id: this.Role
        }));
    }

    RemoveInstanceRule(RuleNumber) {
        // Removes a rule from the instance. (Host & Roles with the permissions can remove rules.)
        this.process(JSON.stringify({
            msg: "InstanceRules",
            mode: "Remove",
            instance_guid: this.Session,
            rule_num: RuleNumber,
            role_id: this.Role
        }));
    }

    UpdateInstanceRule(RuleNumber, Rule_String) {
        // Replaces the old rule with a new one. (Host & Roles with the permissions can update the rules.)
        this.process(JSON.stringify({
            msg: "InstanceRules",
            mode: "Update",
            instance_guid: this.Session,
            rule_num: RuleNumber,
            rule: Rule_String,
            role_id: this.Role
        }));
    }

    GetInstanceRules() {
        // Anyone can request the instance rules if there connected to an instance (might later change that.)
        this.process(JSON.stringify({
            msg: "InstanceRules",
            mode: "Get",
            instance_guid: this.Session
        }));
    }

    CreateRole(RoleName, Color = "#ff0000") {
        // Creates a new role with the color.
        this.process(JSON.stringify({
            msg: "RoleManager",
            instance_guid: this.Session,
            cmd: "CreateRole",
            role_name: RoleName,
            color: Color,
            role_id: this.Role
        }));
    }

    EditRolePermission(PermissionID, RoleName) {
        // Edits the role's permissions set by the instance.
        this.process(JSON.stringify({
            msg: "RoleManager",
            instance_guid: this.Session,
            cmd: "EditRole",
            permissionId: PermissionID,
            role_name: RoleName,
            role_id: this.Role
        }));
    }

    DeleteRole(RoleName) {
        // Deletes the role and removes it entirely.
        this.process(JSON.stringify({
            msg: "RoleManager",
            instance_guid: this.Session,
            cmd: "DeleteRole",
            role_name: RoleName,
            role_id: this.Role
        }));
    }

    AddClientToRole(RoleName, SocketID) {
        // Assigns the role to a socket connected to the session.
        this.process(JSON.stringify({
            msg: "RoleManager",
            instance_guid: this.Session,
            cmd: "AddClientToRole",
            role_name: RoleName,
            socketid: SocketID,
            role_id: this.Role
        }));
    }

    RemoveClientFromRole(RoleName, SocketID) {
        // Un-assign the role from a socket.
        this.process(JSON.stringify({
            msg: "RoleManager",
            instance_guid: this.Session,
            cmd: "RemoveClientFromRole",
            role_name: RoleName,
            socketid: SocketID,
            role_id: this.Role
        }));
    }

    EditRoleColor(RoleName, Color) {
        // Modify the role's color.
        this.process(JSON.stringify({
            msg: "RoleManager",
            instance_guid: this.Session,
            cmd: "EditRoleColor",
            role_name: RoleName,
            color: Color,
            role_id: this.Role
        }));
    }

    ViewRoles() {
        // Views a list of roles.
        this.process(JSON.stringify({
            msg: "ViewRoles",
            instance_guid: this.Session
        }));
    }

    LockRole(RoleName) {
        // Locks the role down to host only.
        this.process(JSON.stringify({
            msg: "LockRole",
            instance_guid: this.Session,
            role_name: RoleName
        }));
    }

    ReportUserToHost(Against, Reason) {
        // Sends a report to the host.
        this.process(JSON.stringify({
            msg: "report",
            instance_guid: this.Session,
            against: Against,
            reason: Reason
        }));
    }

    AboutInstantiated() {
        // Sends an about Instantiated with JSON.
        this.process(JSON.stringify({
            msg: "about"
        }));
    }

    ReportUserToStaff(SocketID, MESSAGE) {
        // Globally reports the user to Instantiated Staff members who have access to the report system.
        this.process(JSON.stringify({
            msg: "FileReport",
            socketid: SocketID,
            Message: MESSAGE
        }));
    }

    GetMOTD() {
        // Grab's instantiated's MOTD message.
        this.process(JSON.stringify({
            msg: "SendMotd"
        }));
    }

    LockInstance() {
        // Locks the instance so no new clients can join.
        this.process(JSON.stringify({
            msg: "toggle_lock",
            instance_guid: this.Session,
            role_id: this.Role
        }));
    }

    TakeControl(Password) {
        // Takes over the instance with a root password.
        this.process(JSON.stringify({
            msg: "take_control",
            instance_guid: this.Session,
            passkey: Password
        }));
    }

    JoinServer(InstanceID, Password, Invitation_Code, AcceptWarning = 0) {
        // Joins the instance server with a password and an invite code (if provided one).
        this.process(JSON.stringify({
            msg: "join_server",
            instance_guid: InstanceID,
            passkey: Password,
            invite_code: Invitation_Code,
            accept: AcceptWarning
        }));
    }

    BanIP(IPAddress, SocketID, Duration, Reason) {
        // Bans the ip address from joining the session. Provided with duration and a reason for the ban.
        this.process(JSON.stringify({
            msg: "ban_ip",
            instance_guid: this.Session,
            duration: Duration,
            banip: IPAddress,
            socketid: SocketID,
            reason: Reason,
            role_id: this.Role
        }));
    }

    UnbanIP(IPAddress) {
        // Completely pardons the ip address and unbans it. Allowing that ip address to rejoin back in.
        this.process(JSON.stringify({
            msg: "unban_ip",
            instance_guid: this.Session,
            unbanip: IPAddress,
            role_id: this.Role
        }));
    }

    AddIPToWhitelist(IPAddress) {
        // Adds an ip address to the whitelist.
        this.process(JSON.stringify({
            msg: "add_whitelist",
            instance_guid: this.Session,
            whitelist_ip: IPAddress,
            role_id: this.Role
        }));
    }

    RemoveIPFromWhitelist(IPAddress) {
        // Removes an ip address from the whitelist.
        this.process(JSON.stringify({
            msg: "del_whitelist",
            instance_guid: this.Session,
            whitelist_ip: IPAddress,
            role_id: this.Role
        }));
    }

    ToggleWhitelist() {
        // Toggles whitelisting.
        this.process(JSON.stringify({
            msg: "toggle_whitelist",
            instance_guid: this.Session,
            role_id: this.Role
        }));
    }

    Kick(SocketID, Reason) {
        // Kicks a socket out of the session.
        this.process(JSON.stringify({
            msg: "kick",
            instance_guid: this.Session,
            socketid: SocketID,
            reason: Reason,
            role_id: this.Role
        }));
    }

    GetSessionData(Item_, Password = null) {
        // Grabs the session's item and reads it. (If password is provided then it is required unless you have the permission in your role.)
        this.process(JSON.stringify({
            msg: "get_storage",
            instance_guid: this.Session,
            Item: Item_,
            passkey: Password,
            role_id: this.Role
        }));
    }

    DeleteSessionItem(Item_, Password = null) {
        // Deletes the session's item from session storage. (If password is provided then it is required unless you have the permission in your role.)
        this.process(JSON.stringify({
            msg: "del_storage",
            instance_guid: this.Session,
            Item: Item_,
            passkey: Password,
            role_id: this.Role
        }));
    }

    AddSessionData(Item_, Data_, Password = null) {
        // Adds/Updates the session's item from session storage to a new value. (If password is provided then it is required unless you have the permission in your role.)
        this.process(JSON.stringify({
            msg: "add_storage",
            instance_guid: this.Session,
            Item: Item_,
            Data: Data_,
            passkey: Password,
            role_id: this.Role
        }));
    }

    InspectSession(Inspection = "All") {
        // Only the host instance can inspect the session. Retrieves the instance's information.
        this.process(JSON.stringify({
            msg: "inspect_session",
            instance_guid: this.Session,
            Inspector: Inspection
        }));
    }

    SearchInstances(ApplicationTitle) {
        // Searches through every session with the provided application title.
        this.process(JSON.stringify({
            msg: "get_sessions",
            Title: ApplicationTitle
        }));
    }

    SearchInstanceOnce(ApplicationTitle) {
        // Finds the first instance it finds.
        this.process(JSON.stringify({
            msg: "auto-discover",
            Title: ApplicationTitle
        }));
    }

    ChangeSessionIcon(ImageURL) {
        // Changes the session's icon to an image url.
        this.process(JSON.stringify({
            msg: "ChangeSessionIcon",
            instance_guid: this.Session,
            image_url: ImageURL
        }));
    }

    GetSessionIcon() {
        // Retrieves the session's icon.
        this.process(JSON.stringify({
            msg: "GetSessionIcon",
            instance_guid: this.Session
        }));
    }

    WarnClient(Reason, SocketID) {
        // Warns the client about their behaviour in the instance. (only the host or the role with the permission can warn users.)
        this.process(JSON.stringify({
            msg: "WarnClient",
            instance_guid: this.Session,
            reason: Reason,
            socketid: SocketID,
            role_id: this.Role
        }));
    }

    SavePremiumSession(RootPassword) {
        // Saves the premium session's data. (Must be premium and bought. Otherwise it won't work.)
        this.process(JSON.stringify({
            msg: "SavePremiumSession",
            instance_guid: this.Session,
            root_passkey: RootPassword
        }));
    }
}

module.exports = InstantiatedSDK;